/*
 * gMapMaker
 * Copyright (C) 2007  Damien Debin
 * http://debin.net/gMapMaker/
 * 
 * This file is part of gMapMaker.
 * 
 * Foobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Collections.Generic;
using System.Text;
using FreeImageAPI;
using System.IO;
using System.Diagnostics;
using System.ComponentModel;

using FIBITMAP = System.UInt32;

namespace gMapMaker
{
  public delegate void ReportProgressFunction(string message, double inc);

  class BigImage
  {
    readonly string topleft;
    readonly string bottomright;
    readonly ReportProgressFunction reportFunction;
    readonly BackgroundWorker worker;
    public bool grayscale;
    public bool dither8bpp;
    public FREE_IMAGE_FORMAT imageFormat = FREE_IMAGE_FORMAT.FIF_PNG;
    public LoadSaveFlags saveflags;
    public int operatingMode;
    int imageHeight = -1;
    int imageWidth = -1;
    TileDownloader td;

    public int ImageHeight
    {
      get
      {
        return imageHeight;
      }
    }
    public int ImageWidth
    {
      get
      {
        return imageWidth;
      }
    }

    public BigImage(string topleft, string bottomright, ReportProgressFunction reportFunction, BackgroundWorker worker)
    {
      this.topleft = topleft;
      this.bottomright = bottomright;
      this.reportFunction = reportFunction;
      this.worker = worker;
    }

    static void FreeImage_OutputMessageFunction(FREE_IMAGE_FORMAT format, string msg)
    {
      Debug.WriteLine("FreeImage: " + msg);
    }

    FIBITMAP LoadImageFromCache(string tileType, GMapTile g)
    {
      FIBITMAP image = 0;

      while (image == 0)
      {
        string src_filename = td.DownloadTile(tileType, g);

        if (worker.CancellationPending)
        {
          break;
        }

        if (!File.Exists(src_filename))
        {
          //blank image white
          image = FreeImage.Allocate(GMapTile.TILE_SIZE, GMapTile.TILE_SIZE, 24, 0, 0, 0);
          if (image == 0)
          {
            throw new InvalidOperationException(Properties.Resources.ExceptionCannotBuildBlankImage);
          }
          if (!FreeImage.Invert(image))
          {
            throw new InvalidOperationException(Properties.Resources.ExceptionCannotBuildBlankImage);
          }
          if (Properties.Settings.Default.SaveBlankImageForMissingTiles)
          {
            if (!FreeImage.Save((tileType == "kh" ? FREE_IMAGE_FORMAT.FIF_JPEG : FREE_IMAGE_FORMAT.FIF_PNG), image, src_filename, (int)(tileType == "kh" ? LoadSaveFlags.JPEG_QUALITYSUPERB : LoadSaveFlags.PNG_DEFAULT)))
            {
              throw new InvalidOperationException(Properties.Resources.ExceptionCannotBuildBlankImage);
            }
          }
        }
        else
        {
          image = FreeImage.Load(FreeImage.GetFileType(src_filename, 0), src_filename, 0);
          if (image == 0)
          {
            File.Delete(src_filename);
            if (Properties.Settings.Default.HaltOnImageFileCorrupted)
            {
              throw new InvalidOperationException(String.Format(Properties.Resources.ExceptionLoadingImage, src_filename));
            }
          }
        }
      }

      return image;
    }

    public void BuildBigImage(string filename)
    {
      FreeImage.SetOutputMessage(FreeImage_OutputMessageFunction);

      int topleftx, toplefty, bottomrightx, bottomrighty;
      uint topleftzoom, bottomrightzoom;

      GMapTile.XYZoom(topleft, out topleftx, out toplefty, out topleftzoom);
      GMapTile.XYZoom(bottomright, out bottomrightx, out bottomrighty, out bottomrightzoom);

      td = new TileDownloader(reportFunction, worker);

      FIBITMAP bigImage = 0, image = 0;

      try
      {
        imageWidth = (bottomrightx - topleftx + 1) * GMapTile.TILE_SIZE;
        imageHeight = (bottomrighty - toplefty + 1) * GMapTile.TILE_SIZE;

        if (imageFormat != FREE_IMAGE_FORMAT.FIF_UNKNOWN)
        {
          reportFunction("Allocating memory for image", 0);

          if ((imageWidth > 65535) || (imageHeight > 65535))
          {
            throw new InvalidOperationException(Properties.Resources.ExceptionImageMemoryAllocation);
          }
          bigImage = FreeImage.Allocate(imageWidth, imageHeight, grayscale ? 8 : 24, 0, 0, 0);
          if (bigImage == 0)
          {
            throw new InvalidOperationException(Properties.Resources.ExceptionImageMemoryAllocation);
          }
        }

        int wholeQuant = (bottomrightx - topleftx + 1) * (bottomrighty - toplefty + 1);
        bool paletteInit = false;

        for (int yy = toplefty; yy <= bottomrighty; yy++)
        {
          for (int xx = topleftx; xx <= bottomrightx; xx++)
          {
            if (worker.CancellationPending)
            {
              return;
            }

            try
            {
              GMapTile g = new GMapTile(xx, yy, topleftzoom);

              reportFunction("(Down)loading image " + g.Quadtree, ((!grayscale && dither8bpp ? 0.0 : 50.0) + 100.0) / (double)wholeQuant);

              switch (operatingMode)
              {
                case 0: //satellite only
                  image = LoadImageFromCache("kh", g);
                  break;
                case 1: //driving map only
                  image = LoadImageFromCache("mt", g);
                  break;
                case 3: //terrain map only
                  image = LoadImageFromCache("pt", g);
                  break;
                case 2: //hybrid
                  image = LoadImageFromCache("kh", g);

                  if (worker.CancellationPending)
                  {
                    return;
                  }

                  FIBITMAP image_alpha = LoadImageFromCache("tt", g);

                  if (FreeImage.GetBPP(image_alpha) != 32)
                  {
                    //sometimes alpha tile from GG is a 2bits image, it means tile is blank
                    break;
                  }

                  if (Properties.Settings.Default.DrivingMapTransparency != 100)
                  {
                    FIBITMAP image_alpha2 = FreeImage.Allocate(GMapTile.TILE_SIZE, GMapTile.TILE_SIZE, 32, 0, 0, 0);
                    if (!FreeImage.Paste(image_alpha2, image_alpha, 0, 0, ((int)Properties.Settings.Default.DrivingMapTransparency * 256) / 100))
                    {
                      //sometimes alpha tile from GG is corrupted, it means tile is blank
                      //we keep tile
                    }
                    FreeImage.Unload(image_alpha);
                    image_alpha = image_alpha2;
                  }

                  FIBITMAP image_result = FreeImage.Composite(image_alpha, false, null, image);
                  FreeImage.Unload(image_alpha);
                  if (image_result != 0)
                  {
                    FreeImage.Unload(image);
                    image = image_result;
                  }
                  else
                  {
                    //sometimes alpha tile from GG is corrupted, it means tile is blank
                    //we keep image
                  }
                  break;
              }

              if (worker.CancellationPending)
              {
                return;
              }

              if (imageFormat == FREE_IMAGE_FORMAT.FIF_UNKNOWN)
              {
                continue;
              }

              if (grayscale)
              {
                //uint image2 = FreeImage.ConvertPalette(bigImage, image);
                FIBITMAP image3 = FreeImage.ConvertTo24Bits(image);
                if (image3 == 0)
                {
                  throw new InvalidOperationException(Properties.Resources.ExceptionCannotConvert24bpp);
                }
                FIBITMAP image2 = FreeImage.ConvertTo8Bits(image3);
                FreeImage.Unload(image3);
                if (image2 == 0)
                {
                  throw new InvalidOperationException(Properties.Resources.ExceptionCannotConvertGrayscale);
                }
                FreeImage.Unload(image);
                image = image2;
                if (!paletteInit)
                {
                  if (!FreeImage.PaletteCopy(image, bigImage))
                  {
                    throw new InvalidOperationException(Properties.Resources.ExceptionGrayscalePalette);
                  }
                  paletteInit = true;
                }
              }

              reportFunction("Pasting " + g.Quadtree, 0);

              if (!FreeImage.Paste(bigImage, image, (xx - topleftx) * GMapTile.TILE_SIZE, (yy - toplefty) * GMapTile.TILE_SIZE, 256))
              {
                throw new InvalidOperationException(Properties.Resources.ExceptionPastingImage);
              }
            }
            finally
            {
              FreeImage.Unload(image);
              image = 0;
            }
          }
        }

        if (worker.CancellationPending || (imageFormat == FREE_IMAGE_FORMAT.FIF_UNKNOWN))
        {
          return;
        }

        if (!grayscale && dither8bpp)
        {
          reportFunction("Converting image to 8bpp", 0.0);
          uint bigImage8 = FreeImage.ColorQuantize(bigImage, FREE_IMAGE_QUANTIZE.FIQ_WUQUANT);
          if (bigImage8 == 0)
          {
            throw new InvalidOperationException(Properties.Resources.ExceptionConversion24bpp8bpp);
          }
          FreeImage.Unload(bigImage);
          bigImage = bigImage8;
          reportFunction("Image converted", 50.0);
        }

        if (worker.CancellationPending)
        {
          return;
        }

        reportFunction("Saving image to " + Path.GetFileName(filename), 0.0);
        if (!FreeImage.Save(imageFormat, bigImage, filename, (int)saveflags))
        {
          throw new IOException(Properties.Resources.ExceptionCannotSaveFile);
        }

        reportFunction("Image saved", 50.0);
      }
      finally
      {
        FreeImage.Unload(bigImage);
      }
    }
  }
}
